
%    Sparse Coding Classification Code -- (c) 2009-2011 Simon Scholler,
%    Music Technology Group, UPF, Barcelona,
%    and Bernstein Center of Computational Neuroscience, Berlin
%
%    This code is provided as it is, with no guarantees.  Published reports of research
%    using this code (or a modified version, maintaining a significant portion of the original code)
%    should cite the following article:
%
%    Scholler, S., Purwins, H., Sparse Approximations for Drum Sound Classification,
%    IEEE Journal of Selected Topics in Signal Processing, Vol. 5, No. 5:pp. 933-940, 2011 
%
%    Comments and bug reports are welcome.  Email to
%    simon.scholler@googlemail.com.
%    I would also appreciate hearing about how you used this code,
%
%    You are free to use this code and
%    to modify, extend or distribute this code, as long
%    as this copyright notice is included whole and unchanged.
%



function book = mpd(x, dict, opt)
%
%  USAGE:
%  book = function mp_decomp(x, dict)
%
%  IN:   x                -      signal vector
%        dict             -      dictionary of atoms (cell or matrix)
%        opt              -      settings structure (cf. set_mp_opts)
%
%  OUT:  book             -      book of selected atoms 
%

if ~isColumnVec(x) 
   x = x';
end
norm_x = l2norm(x);
len_x = length(x);

if ~iscell(dict)
    matdict = dict;
    dict = cell(1,size(dict,2));
    for n = 1:length(dict)
        dict{n} = matdict(:,n);
    end
end
%maxlen = max(cellfun(@(y) length(y),dict));
nDictAtoms = length(dict);

if opt.normalize_dict
   dict = dict_normalize(dict);
end

%% Decompose the Signal
book.cX = {}; book.X = []; book.w = []; book.idx = []; book.SNR = [];
book.shift = []; book.nAtoms = 0; book.residual = x;
book.Corr = cell(1,length(dict)); book.lastup_startend = [1 len_x];

maxcorr = zeros(1,nDictAtoms);
shift = zeros(1,nDictAtoms);
if strcmpi(opt.stop,'SNR')
   snr = 0;
   while snr<opt.SNR && length(book.w)<opt.maxAtoms
      mp();
      snr = signal_energy(book.sigmodel)/signal_energy(book.residual);
   end
elseif strcmpi(opt.stop,'atoms')
   for n = 1:opt.nAtoms      
      mp();            
   end
elseif strcmpi(opt.stop,'corr')
    ct = 1;
    while ct>=opt.corrthresh
        ct = mp();
    end
else
    error('stopping criterion unknown')
end

book.X = book.X(:,1:book.nAtoms);
book.sigmodel = book.X*book.w;
book.residual = x-book.sigmodel;
book.snr = signal_energy(book.sigmodel)/signal_energy(book.residual);




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Matching Pursuit Subfunction
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   function ct = mp()
      % calculate one matching pursuit iteration
      sgn = zeros(1,nDictAtoms);
      for a = 1:nDictAtoms
          %c = conv1(book.residual,dict{a});
          i = book.lastup_startend(1); j = book.lastup_startend(2);
          la = length(dict{a});
          if isempty(book.Corr{a})
              book.Corr{a} = zeros(len_x-la+1,1);
              j = len_x-la+1;
          end
          book.Corr{a}(max(1,i-la+1):min(j,len_x-la+1)) = conv1(book.residual(max(1,i-la+1):min(len_x,j+la-1)),dict{a});
          if ~opt.reuse_atom && any(a==book.idx)
              maxcorr(a) = 0;
          else
              [maxcorr(a) shift(a)] = max(abs(book.Corr{a}));
              sgn(a) = sign(book.Corr{a}(shift(a)));
          end
      end
      [w, idx] = max(maxcorr);   % find which basis function gave highest dotproduct
    
      
      %%% Update the representation basis %%%
      atom = dict{idx};
      book.cX{end+1} = atom;
      atomX = zeros(len_x,1);
      atomX(shift(idx):shift(idx)+length(atom)-1) = atom;      
      book.lastup_startend = [shift(idx) shift(idx)+length(atom)-1];         
      if isempty(book.nAtoms) || mod(book.nAtoms,500)==0
          book.X = [book.X zeros(len_x,500)];
      end
      book.X(:,book.nAtoms+1) = atomX;    % add atom to book
      book.nAtoms = book.nAtoms+1;
      book.w = [book.w; w*sgn(idx);];   
      ct = abs(book.w(end)); 
      book.idx = [book.idx idx];
      book.shift = [book.shift shift(idx)];            
                  
      if mod(book.nAtoms,opt.snr_every)==0
          book.sigmodel = book.X(:,1:book.nAtoms)*book.w;      
          book.residual = x - book.sigmodel;
          snr = signal_energy(book.sigmodel)/signal_energy(book.residual);
          book.SNR = [book.SNR snr];
      else
          book.residual = book.residual - book.w(end)*atomX;
      end          
   end

end
         